"""DID Document and resource builders."""

from typing import Iterator, List, Optional, Type, Union

from ..did import DID
from ..did_url import DIDUrl
from ..service import DIDCommService, Service
from ..verification_method import VerificationMethod
from .doc import DIDDocument


def _default_id_generator(base: str, start: int = 0) -> Iterator[str]:
    """Generate ID fragments."""
    index = start
    while True:
        yield "{}-{}".format(base, index)
        index += 1


class VerificationMethodBuilder:
    """VerificationMethod scoped builder."""

    def __init__(
        self,
        did: DID,
        *,
        id_base: str = None,
        methods: Optional[List[VerificationMethod]] = None,
    ):
        """Initialize builder."""
        self._did = did
        self.methods = methods or []
        self._id_base = id_base or "key"
        self._id_generator = self._default_id_generator()

    def _default_id_generator(self):
        """Default ID generator."""
        yield from _default_id_generator(self._id_base, start=len(self.methods))

    def add(
        self,
        type_: Type[VerificationMethod],
        ident: Optional[str] = None,
        controller: DID = None,
        **kwargs,
    ):
        """Add verification method from parts and context."""
        ident = ident or next(self._id_generator)
        controller = controller or self._did
        vmethod = type_.make(id=self._did.ref(ident), controller=controller, **kwargs)
        self.methods.append(vmethod)
        return vmethod

    def remove(self, vmethod: VerificationMethod):
        """Remove method from builder."""
        self.methods.remove(vmethod)


class RelationshipBuilder(VerificationMethodBuilder):
    """Builder for relationships."""

    def __init__(
        self,
        did: DID,
        id_base: str,
        *,
        methods: Optional[List[Union[VerificationMethod, DIDUrl]]] = None,
    ):
        """Initialize builder."""
        super().__init__(did, id_base=id_base)
        self.methods = methods or []

    def _default_id_generator(self):
        """Default ID generator."""
        start = len(
            [
                vmethod
                for vmethod in self.methods
                if isinstance(vmethod, VerificationMethod)
            ]
        )
        yield from _default_id_generator(self._id_base, start=start)

    def reference(self, ref: DIDUrl):
        """Add reference to relationship."""
        if not isinstance(ref, DIDUrl):
            raise ValueError(
                "Reference must be DIDUrl, not {}".format(type(ref).__name__)
            )

        self.methods.append(ref)

    def embed(self, *args, **kwargs):
        """Embed verification method in relationship."""
        return super().add(*args, **kwargs)

    def remove(self, vmethod: Union[DIDUrl, VerificationMethod]):
        """Remove reference or method from builder."""
        self.methods.remove(vmethod)


class ServiceBuilder:
    """Builder for services."""

    def __init__(self, did: DID, *, services: Optional[List[Service]] = None):
        """Initialize builder."""
        self._did = did
        self.services = services or []
        self._id_generator = _default_id_generator("service", start=len(self.services))

    def _determine_next_priority(self):
        """Return the next priority after the highest priority currently in services."""
        return (
            max(
                [
                    service.priority
                    for service in self.services
                    if isinstance(service, DIDCommService)
                ]
            )
            + 1
            if self.services
            else 0
        )

    def add(
        self, type_: str, service_endpoint: str, ident: Optional[str] = None, **extra
    ):
        """Add service."""
        ident = ident or next(self._id_generator)
        service = Service.make(
            id=self._did.ref(ident),
            type=type_,
            service_endpoint=service_endpoint,
            **extra,
        )
        self.services.append(service)
        return service

    def add_didcomm(
        self,
        service_endpoint: str,
        recipient_keys: List[Union[VerificationMethod, DIDUrl, str]],
        routing_keys: Optional[List[Union[VerificationMethod, DIDUrl, str]]] = None,
        *,
        priority: Optional[int] = None,
        type_: Optional[str] = None,
        ident: Optional[str] = None,
        accept: Optional[List[str]] = None,
    ):
        """Add DIDComm Service."""
        ident = ident or next(self._id_generator)
        routing_keys = routing_keys or []
        priority = priority or self._determine_next_priority()
        service = DIDCommService.make(
            id=self._did.ref(ident),
            service_endpoint=service_endpoint,
            recipient_keys=[
                vmethod.id if isinstance(vmethod, VerificationMethod) else vmethod
                for vmethod in recipient_keys
            ],
            routing_keys=[
                vmethod.id if isinstance(vmethod, VerificationMethod) else vmethod
                for vmethod in routing_keys
            ],
            type=type_,
            priority=priority,
            accept=accept,
        )
        self.services.append(service)
        return service

    def remove(self, service: Service):
        """Remove service from builder."""
        self.services.remove(service)


class DIDDocumentBuilder:
    """Builder for constructing DID Documents programmatically."""

    def __init__(
        self,
        id: Union[str, DID],
        context: List[str] = None,
        *,
        also_known_as: List[str] = None,
        controller: Union[List[str], List[DID]] = None,
    ):
        """Initliaze builder."""
        self.id: DID = DID(id)
        self.context = context or self.__default_context()
        self.also_known_as = also_known_as
        self.controller = controller
        self.verification_method = VerificationMethodBuilder(self.id)
        self.authentication = RelationshipBuilder(self.id, "auth")
        self.assertion_method = RelationshipBuilder(self.id, "assert")
        self.key_agreement = RelationshipBuilder(self.id, "key-agreement")
        self.capability_invocation = RelationshipBuilder(self.id, "capability-invocation")
        self.capability_delegation = RelationshipBuilder(self.id, "capability-delegation")
        self.service = ServiceBuilder(self.id)
        self.extra = {}

    @staticmethod
    def __default_context() -> List[str]:
        return ["https://www.w3.org/ns/did/v1"]

    @classmethod
    def from_doc(cls, doc: DIDDocument) -> "DIDDocumentBuilder":
        """Create a Builder from an existing DIDDocument."""
        builder = cls(
            id=doc.id,
            context=doc.context,
            also_known_as=doc.also_known_as,
            controller=doc.controller,
        )
        builder.verification_method = VerificationMethodBuilder(
            doc.id, methods=doc.verification_method
        )
        builder.authentication = RelationshipBuilder(
            doc.id, "auth", methods=doc.authentication
        )
        builder.assertion_method = RelationshipBuilder(
            doc.id, "assert", methods=doc.assertion_method
        )
        builder.key_agreement = RelationshipBuilder(
            doc.id, "key-agreement", methods=doc.key_agreement
        )
        builder.capability_invocation = RelationshipBuilder(
            doc.id, "capability-invocation", methods=doc.capability_invocation
        )
        builder.capability_delegation = RelationshipBuilder(
            doc.id, "capability-delegation", methods=doc.capability_delegation
        )
        builder.service = ServiceBuilder(doc.id, services=doc.service)
        return builder

    def build(self) -> DIDDocument:
        """Build document."""
        return DIDDocument.model_construct(
            id=self.id,
            context=self.context,
            also_known_as=self.also_known_as,
            controller=self.controller,
            verification_method=self.verification_method.methods or None,
            authentication=self.authentication.methods or None,
            assertion_method=self.assertion_method.methods or None,
            key_agreement=self.key_agreement.methods or None,
            capability_invocation=self.capability_invocation.methods or None,
            capability_delegation=self.capability_delegation.methods or None,
            service=self.service.services or None,
            **self.extra,
        )



"""Resource class that forms the base of all DID Document components."""

import json
from abc import ABC, abstractmethod
from typing import Any, Dict, Type, TypeVar

import typing_extensions
from pydantic import BaseModel, ConfigDict, TypeAdapter, alias_generators
from typing_extensions import Literal

from .validation import wrap_validation_error

ResourceType = TypeVar("ResourceType", bound="Resource")


if hasattr(typing_extensions, "get_args"):  # pragma: no cover
    from typing_extensions import get_args, get_origin

    def get_literal_values(literal):
        """Return the args of a literal."""
        return get_args(literal)

    def is_literal(type_):
        """Return if type is literal."""
        return get_origin(type_) is Literal

else:  # pragma: no cover
    # Python 3.6 and Literals behave differently
    from typing_extensions import _Literal

    def get_literal_values(literal):
        """Return the args of a literal."""
        return literal.__values__

    def is_literal(type_):
        """Return if type is literal."""
        return isinstance(type_, _Literal)


class Resource(BaseModel):
    """Base class for DID Document components."""

    model_config = ConfigDict(
        populate_by_name=True,
        extra="allow",
        alias_generator=alias_generators.to_camel,
    )

    def serialize(self):
        """Return serialized representation of Resource."""
        return self.model_dump(exclude_none=True, by_alias=True)

    @classmethod
    def deserialize(cls: Type[ResourceType], value: dict) -> ResourceType:
        """Deserialize into Resource subtype."""
        with wrap_validation_error(
            ValueError,
            message=f"Failed to deserialize {cls.__name__}",
        ):
            resource_adapter = TypeAdapter(cls)
            return resource_adapter.validate_python(value)

    @classmethod
    def from_json(cls, value: str):
        """Deserialize Resource from JSON."""
        loaded: dict = json.loads(value)
        return cls.deserialize(loaded)

    def to_json(self):
        """Serialize Resource to JSON."""
        return self.model_dump_json(exclude_none=True, by_alias=True)

    @classmethod
    def _fill_in_required_literals(cls, **kwargs) -> Dict[str, Any]:
        """Return dictionary of field name to value from literals."""
        for field in cls.model_fields.values():
            field_name = field.alias
            field_type = field.annotation
            if (
                field.is_required()
                and is_literal(field_type)
                and (field_name not in kwargs or kwargs[field_name] is None)
            ):
                kwargs[field_name] = get_literal_values(field_type)[0]
        return kwargs

    @classmethod
    def _overwrite_none_with_defaults(cls, **kwargs) -> Dict[str, Any]:
        """Overwrite none values in kwargs with defaults for corresponding field."""
        for field in cls.model_fields.values():
            field_name = field.alias
            if field_name in kwargs and kwargs[field_name] is None:
                kwargs[field_name] = field.get_default()
        return kwargs

    @classmethod
    def make(cls: Type[ResourceType], **kwargs) -> ResourceType:
        """Create instance of class, filling in literals."""
        kwargs = cls._fill_in_required_literals(**kwargs)
        kwargs = cls._overwrite_none_with_defaults(**kwargs)
        return cls(**kwargs)


class IndexedResource(Resource, ABC):
    """Resource with index for supporting dereferencing nested objects."""

    _index: dict = {}

    def __init__(self, **data):
        """Initialize Resource."""
        super().__init__(**data)
        self._index_resources()

    @abstractmethod
    def _index_resources(self):
        """Index nested resources."""

    @abstractmethod
    def dereference(self, reference: str) -> Resource:
        """Dereference a nested object."""

    def dereference_as(self, typ: Type[ResourceType], reference: str) -> ResourceType:
        """Dereference a resource to a specific type."""
        with wrap_validation_error(
            ValueError,
            message=f"Dereferenced resource {reference} could not be parsed as {typ}",
        ):
            resource = self.dereference(reference)
            resource_adapter: TypeAdapter[ResourceType] = TypeAdapter(typ)
            return resource_adapter.validate_python(resource.model_dump())

    @classmethod
    def model_construct(cls, **data):
        """Construct and index."""
        resource = super(Resource, cls).model_construct(**data)
        resource._index_resources()
        return resource



"""DID Doc Verification Method."""

from typing import ClassVar, Optional, Set, Type, Union

from inflection import underscore
from pydantic import create_model, field_validator, model_validator, alias_generators
from typing_extensions import Literal

from pydid.validation import required_group

from .did import DID
from .did_url import DIDUrl, InvalidDIDUrlError
from .resource import Resource

set_of_material_properties = {
    "blockchain_account_id",
    "ethereum_address",
    "public_key_base58",
    "public_key_gpg",
    "public_key_hex",
    "public_key_jwk",
    "public_key_pem",
    "public_key_multibase",
}
material_properties_camel = {
    alias_generators.to_camel(prop) for prop in set_of_material_properties
}


class VerificationMaterial:
    """Type annotation marker for the material attribute of a verification method."""


class VerificationMaterialUnknown(NotImplementedError):
    """Raised when verification material is unknown but access is attempted."""


class VerificationMethod(Resource):
    """Representation of DID Document Verification Methods."""

    material_properties: ClassVar[Set[str]] = set_of_material_properties

    id: DIDUrl
    type: str
    controller: DID
    public_key_hex: Optional[str] = None
    public_key_base58: Optional[str] = None
    public_key_pem: Optional[str] = None
    public_key_multibase: Optional[str] = None
    blockchain_account_id: Optional[str] = None
    ethereum_address: Optional[str] = None
    public_key_jwk: Optional[dict] = None
    _material_prop: Optional[str] = None

    def __init__(self, **data):
        """Initialize a VerificationMethod."""
        super().__init__(**data)
        self._material_prop = self._material_prop or self._infer_material_prop()

    @classmethod
    def suite(cls: Type, typ: str, material: str, material_type: Type):
        """Return a subclass of VerificationMethod for a given type."""
        model = create_model(
            typ,
            __module__=cls.__module__,
            __base__=cls,
            type=(Literal[typ], ...),  # type:ignore
            **{underscore(material): (material_type, ...)},
        )
        model.material = property(
            lambda self: getattr(self, underscore(material)),
            lambda self, value: setattr(self, underscore(material), value),
        )
        model._material_prop = underscore(material)
        return model

    @field_validator("type", mode="before")
    @classmethod
    def _allow_type_list(cls, value: Union[str, list[str]]) -> str:
        """Unwrap type list.

        This validator handles a common DID Document mutation.
        """
        if isinstance(value, list):
            return value[0]
        return value

    @field_validator("controller", mode="before")
    @classmethod
    def _allow_controller_list(cls, value: Union[DID, list[DID]]) -> DID:
        """Unwrap controller list.

        This validator handles a common DID Document mutation.
        """
        if isinstance(value, list):
            return value[0]
        return value

    @model_validator(mode="before")
    @classmethod
    def _allow_missing_controller(cls, values: Union[dict, "VerificationMethod"]):
        """Derive controller value from ID.

        This validator handles a common DID Document mutation.
        """
        if not isinstance(values, dict):
            values = values.__dict__

        if "controller" not in values:
            if "id" not in values:
                raise ValueError(
                    "Could not derive controller: no id present in verification method"
                )
            try:
                ident = DIDUrl.parse(values["id"])
                values["controller"] = ident.did
            except InvalidDIDUrlError:
                pass
        return values

    @model_validator(mode="before")
    @classmethod
    def _method_appears_to_contain_material(
        cls, values: Union[dict, "VerificationMethod"]
    ):
        """Validate that the method appears to contain verification material."""
        if not isinstance(values, dict):
            values = values.__dict__

        if values.get("controller") and len(values) < 4:
            raise ValueError(
                "Key material expected, found: {}".format(list(values.keys()))
            )
        return values

    @model_validator(mode="before")
    @classmethod
    def _no_more_than_one_material_prop(cls, values: Union[dict, "VerificationMethod"]):
        """Validate that exactly one material property was specified on method."""
        if not isinstance(values, dict):
            values = values.__dict__

        model_properties = {key for key, value in values.items() if value is not None}

        set_material_properties = set_of_material_properties & model_properties
        set_material_properties_camel = material_properties_camel & model_properties

        if len(set_material_properties | set_material_properties_camel) > 1:
            raise ValueError(
                "Found properties {}; only one is allowed".format(
                    ", ".join(set_material_properties)
                )
            )
        return values

    def _infer_material_prop(self) -> Optional[str]:
        """Guess the property that appears to be the verification material.

        Guess is based on known material property names.
        """

        for prop, value in self:
            if prop in self.material_properties and value is not None:
                return prop

        return None

    @property
    def material(self):
        """Return material."""
        if not self._material_prop:
            raise VerificationMaterialUnknown(
                "Verification Material is not known for this method"
            )
        return getattr(self, self._material_prop)

    @classmethod
    def method_type(cls) -> Optional[str]:
        """Return method type if known."""
        return cls._fill_in_required_literals().get("type")


# Verification Method Suites registered in DID Spec


class Ed25519VerificationKey2018(VerificationMethod):
    """Ed25519VerificationKey2018 VerificationMethod."""

    type: Literal["Ed25519VerificationKey2018"]
    public_key_base58: str


class Ed25519VerificationKey2020(VerificationMethod):
    """Ed25519VerificationKey2020 VerificationMethod."""

    type: Literal["Ed25519VerificationKey2020"]
    public_key_multibase: str


class OpenPgpVerificationKey2019(VerificationMethod):
    """OpenPgpVerificationKey2019 VerificationMethod."""

    type: Literal["OpenPgpVerificationKey2019"]
    public_key_pem: str


class JsonWebKey2020(VerificationMethod):
    """JsonWebKey2020 VerificationMethod."""

    type: Literal["JsonWebKey2020"]
    public_key_jwk: dict


class EcdsaSecp256k1VerificationKey2019(VerificationMethod):
    """EcdsaSecp256k1VerificationKey2019 VerificationMethod."""

    type: Literal["EcdsaSecp256k1VerificationKey2019"]
    _require_one_of = required_group({"public_key_jwk", "public_key_hex"})


class Bls1238G1Key2020(VerificationMethod):
    """Bls1238G1Key2020 VerificationMethod."""

    type: Literal["Bls1238G1Key2020"]
    public_key_base58: str


class Bls1238G2Key2020(VerificationMethod):
    """Bls1238G2Key2020 VerificationMethod."""

    type: Literal["Bls1238G2Key2020"]
    public_key_base58: str


class GpgVerificationKey2020(VerificationMethod):
    """GpgVerificationKey2020 VerificationMethod."""

    type: Literal["GpgVerificationKey2020"]
    public_key_gpg: str


class RsaVerificationKey2018(VerificationMethod):
    """RsaVerificationKey2018 VerificationMethod."""

    type: Literal["RsaVerificationKey2018"]
    public_key_jwk: dict


class X25519KeyAgreementKey2019(VerificationMethod):
    """X25519KeyAgreementKey2019 VerificationMethod."""

    type: Literal["X25519KeyAgreementKey2019"]
    public_key_base58: str


class X25519KeyAgreementKey2020(VerificationMethod):
    """X25519KeyAgreementKey2020 VerificationMethod."""

    type: Literal["X25519KeyAgreementKey2020"]
    public_key_multibase: str


class SchnorrSecp256k1VerificationKey2019(VerificationMethod):
    """SchnorrSecp256k1VerificationKey2019 VerificationMethod."""

    type: Literal["SchnorrSecp256k1VerificationKey2019"]


class EcdsaSecp256k1RecoveryMethod2020(VerificationMethod):
    """EcdsaSecp256k1RecoveryMethod2020 VerificationMethod."""

    type: Literal["EcdsaSecp256k1RecoveryMethod2020"]
    _require_one_of = required_group(
        {"public_key_jwk", "public_key_hex", "ethereum_address", "blockchain_account_id"}
    )


class Multikey(VerificationMethod):
    """Multikey VerificationMethod."""

    type: Literal["Multikey"]
    public_key_multibase: str


class UnsupportedVerificationMethod(VerificationMethod):
    """Model representing unsupported verification methods."""


class UnknownVerificationMethod(VerificationMethod):
    """Model representing unknown verification methods."""


KnownVerificationMethods = Union[
    Ed25519VerificationKey2018,
    Ed25519VerificationKey2020,
    OpenPgpVerificationKey2019,
    JsonWebKey2020,
    EcdsaSecp256k1VerificationKey2019,
    Bls1238G1Key2020,
    Bls1238G2Key2020,
    GpgVerificationKey2020,
    RsaVerificationKey2018,
    X25519KeyAgreementKey2019,
    X25519KeyAgreementKey2020,
    SchnorrSecp256k1VerificationKey2019,
    EcdsaSecp256k1RecoveryMethod2020,
    Multikey,
]



"""Common DID Document corrections."""

from typing import Union

from ..did import DID


def insert_missing_ids(value: dict) -> dict:
    """Insert missing resource IDs.

    This correction can be applied directly to a raw document dictionary or
    with the pydid.deserialize_document method:

    >>> import pydid
    >>> doc_raw = {"@context": "http://example.com", "id": "did:example:123"}
    >>> doc_raw = pydid.corrections.insert_missing_ids(doc_raw)

    Or:
    >>> import pydid
    >>> doc_raw = {"@context": "http://example.com", "id": "did:example:123"}
    >>> doc = pydid.deserialize_document(
    ...     doc_raw, corrections=[pydid.corrections.insert_missing_ids]
    ... )
    """
    if "id" not in value:
        raise ValueError("No ID found in Document.")

    did = DID(value["id"])
    index = 0

    def _walk(value: Union[str, list, dict]):
        nonlocal index
        if isinstance(value, str):
            return
        if isinstance(value, list):
            for item in value:
                _walk(item)
            return
        if isinstance(value, dict):
            if "id" not in value:
                value["id"] = str(did.ref("inserted-{}".format(index)))
                index += 1
                return

    for nested in value.values():
        _walk(nested)

    return value


def public_key_is_verification_method(value: dict) -> dict:
    """Transform publicKey to verificationMethod.

    This is helpful if dealing with DID resolvers that have not updated to the
    latest DID Spec.

    This correction can be applied directly to a raw document dictionary or
    with the pydid.deserialize_document method:

    >>> import pydid
    >>> doc_raw = {
    ...     "@context": "http://example.com",
    ...     "id": "did:example:123",
    ...     "publicKey": [{"id": "did:example:123#key-1"}],
    ... }
    >>> doc_raw = pydid.corrections.public_key_is_verification_method(doc_raw)
    >>> assert "verificationMethod" in doc_raw
    >>> assert "publicKey" not in doc_raw

    Or:
    >>> import pydid
    >>> doc_raw = {
    ...     "@context": "http://example.com",
    ...     "id": "did:example:123",
    ...     "publicKey": [{"id": "did:example:123#key-1"}],
    ... }
    >>> doc = pydid.deserialize_document(
    ...     doc_raw, corrections=[pydid.corrections.public_key_is_verification_method]
    ... )
    >>> assert doc.verification_method
    """

    if "publicKey" in value:
        value["verificationMethod"] = value.pop("publicKey")
    return value



"""Validation tools and helpers."""

from contextlib import contextmanager
from typing import Set, Type

from pydantic import ValidationError, ValidationInfo, model_validator


@contextmanager
def wrap_validation_error(error_to_raise: Type[Exception], message: str = None):
    """Wrap validation errors with more friendly errors."""
    try:
        yield
    except ValidationError as error:
        raise error_to_raise(
            ":\n".join([message, str(error)]) if message else str(error)
        ) from error


def required_group(props: Set[str]):
    """Require at least one of the properties to be present."""

    def _require_group(_model, _: ValidationInfo):
        if not isinstance(_model, dict):
            _model = _model.__dict__

        defined_props = props & {
            key for key, value in _model.items() if value is not None
        }
        if len(defined_props) < 1:
            raise ValueError("At least one of {} was required; none found".format(props))
        return _model

    return model_validator(mode="after")(_require_group)



"""W3C DID Representation.

DID Parsing rules derived from W3C Decentralized Identifiers v1.0 Working Draft 18
January 2021:

    https://w3c.github.io/did-core/

"""

from typing import Any, Dict, Optional

from pydantic import GetCoreSchemaHandler, GetJsonSchemaHandler
from pydantic.json_schema import JsonSchemaValue
from pydantic_core import CoreSchema, core_schema

from .common import DID_PATTERN, DIDError
from .did_url import DIDUrl


class InvalidDIDError(DIDError, ValueError):
    """Invalid DID."""


class DID(str):
    """DID Representation and helpers."""

    def __init__(self, did: str):
        """Validate and parse raw DID str."""
        super().__init__()
        if isinstance(did, DID):
            self._method = did.method
            self._id = did._id
            return

        matched = DID_PATTERN.match(did)
        if not matched:
            raise InvalidDIDError("Unable to parse DID {}".format(did))
        self._method = matched.group(1)
        self._id = matched.group(2)

    @classmethod
    def __get_pydantic_core_schema__(
        cls, source_type: Any, handler: GetCoreSchemaHandler
    ) -> CoreSchema:
        """Get core schema."""
        return core_schema.no_info_after_validator_function(cls, handler(str))

    @classmethod
    def __get_pydantic_json_schema__(
        cls, core_schema: CoreSchema, handler: GetJsonSchemaHandler
    ) -> JsonSchemaValue:
        """Update schema fields."""
        json_schema = handler(core_schema)
        json_schema["pattern"] = DID_PATTERN
        return json_schema

    @property
    def method(self):
        """Return the method of this DID."""
        return self._method

    @property
    def method_specific_id(self):
        """Return the method specific identifier."""
        return self._id

    def url(
        self,
        path: Optional[str] = None,
        query: Optional[Dict[str, str]] = None,
        fragment: Optional[str] = None,
    ):
        """Return a DID URL for this DID."""
        return DIDUrl.unparse(self, path, query, fragment)

    def ref(self, ident: str) -> DIDUrl:
        """Return a DID reference (URL) for use as an ID in a DID Doc section."""
        return DIDUrl.unparse(self, fragment=ident)

    @classmethod
    def is_valid(cls, did: str):
        """Return if the passed string is a valid DID."""
        return DID_PATTERN.match(did)

    @classmethod
    def model_validate(cls, did: str):
        """Validate the given string as a DID."""
        if not cls.is_valid(did):
            raise InvalidDIDError('"{}" is not a valid DID'.format(did))
        return did

    @classmethod
    def validate(cls, did: str):
        """Validate the given string as a DID."""
        return cls.model_validate(did)

    @classmethod
    def _validate(cls, did):
        """Pydantic validator."""
        return cls(did)



"""DID Document Object."""

from abc import ABC
from typing import Any, List, Optional, Union

from pydantic import Field, field_validator
from typing_extensions import Annotated

from ..did import DID, InvalidDIDError
from ..did_url import DIDUrl, InvalidDIDUrlError
from ..resource import IndexedResource, Resource
from ..service import DIDCommV1Service, DIDCommV2Service, Service
from ..verification_method import (
    KnownVerificationMethods,
    UnknownVerificationMethod,
    VerificationMethod,
)


class DIDDocumentError(Exception):
    """Base Exception for problems with DID Documents."""


class IdentifiedResourceMismatch(DIDDocumentError):
    """Raised when two or more of the same ID point to differing resources."""


class IDNotFoundError(DIDDocumentError):
    """Raised when Resource ID not found in DID Document."""


class DIDDocumentRoot(Resource):
    """Representation of DID Document."""

    context: Annotated[List[Union[str, dict]], Field(alias="@context")] = [
        "https://www.w3.org/ns/did/v1"
    ]  # noqa: F722
    id: DID
    also_known_as: Optional[List[str]] = None
    controller: Optional[List[DID]] = None
    verification_method: Optional[List[VerificationMethod]] = None
    authentication: Optional[List[Union[DIDUrl, VerificationMethod]]] = None
    assertion_method: Optional[List[Union[DIDUrl, VerificationMethod]]] = None
    key_agreement: Optional[List[Union[DIDUrl, VerificationMethod]]] = None
    capability_invocation: Optional[List[Union[DIDUrl, VerificationMethod]]] = None
    capability_delegation: Optional[List[Union[DIDUrl, VerificationMethod]]] = None
    service: Optional[List[Service]] = None

    @field_validator("context", "controller", mode="before")
    @classmethod
    def _listify(cls, value) -> Optional[list]:
        """Transform values into lists that are allowed to be a list or single."""
        if value is None:
            return
        if isinstance(value, list):
            return value
        return [value]


class BaseDIDDocument(DIDDocumentRoot, IndexedResource, ABC):
    """Abstract BaseDIDDocument class."""

    @property
    def is_nonconformant(self):
        """Return whether doc is non-conformant."""
        return isinstance(self, NonconformantDocument)

    @property
    def is_conformant(self):
        """Return whether doc is conformant."""
        return not isinstance(self, NonconformantDocument)


class BasicDIDDocument(BaseDIDDocument):
    """Basic DID Document."""

    def _index_resources(self):
        """Index resources by ID.

        IDs are not guaranteed to be unique within the document.
        The first instance is stored in the index and subsequent id collisions
        are checked against the original. If they do not match, an error will
        be thrown.
        """

        def _indexer(item):
            if not item:
                # Attribute isn't set
                return
            if isinstance(item, DIDUrl):
                # We don't index references
                return
            if isinstance(item, list):
                for subitem in item:
                    _indexer(subitem)
                return

            assert isinstance(item, (VerificationMethod, Service))
            if item.id in self._index and item != self._index[item.id]:
                raise IdentifiedResourceMismatch(
                    "ID {} already found in Index and Items do not match".format(item.id)
                )

            if not item.id.did:
                key = item.id.as_absolute(self.id)
            else:
                key = item.id

            self._index[key] = item

        for item in (
            self.verification_method,
            self.authentication,
            self.assertion_method,
            self.key_agreement,
            self.capability_invocation,
            self.capability_delegation,
            self.service,
        ):
            _indexer(item)

    def dereference(self, reference: Union[str, DIDUrl]) -> Resource:
        """Dereference a DID URL to a document resource."""
        if isinstance(reference, str):
            reference = DIDUrl.parse(reference)
        if not reference.did:
            reference = reference.as_absolute(self.id)

        if reference not in self._index:
            raise IDNotFoundError("ID {} not found in document".format(reference))
        return self._index[reference]


PossibleMethodTypes = Union[KnownVerificationMethods, UnknownVerificationMethod]
PossibleServiceTypes = Union[DIDCommV1Service, DIDCommV2Service, Service]


class DIDDocument(BasicDIDDocument):
    """DID Document for DID Spec version 1.0.

    Registered verification method and service types are parsed into specific objects.
    """

    verification_method: Optional[List[PossibleMethodTypes]] = None
    authentication: Optional[List[Union[DIDUrl, PossibleMethodTypes]]] = None
    assertion_method: Optional[List[Union[DIDUrl, PossibleMethodTypes]]] = None
    key_agreement: Optional[List[Union[DIDUrl, PossibleMethodTypes]]] = None
    capability_invocation: Optional[List[Union[DIDUrl, PossibleMethodTypes]]] = None
    capability_delegation: Optional[List[Union[DIDUrl, PossibleMethodTypes]]] = None
    service: Optional[List[PossibleServiceTypes]] = None

    @classmethod
    def deserialize(cls, value: dict) -> "DIDDocument":
        """Wrap deserialization with a basic validation pass before matching to type."""
        DIDDocumentRoot.deserialize(value)
        return super(DIDDocument, cls).deserialize(value)


class NonconformantDocument(BaseDIDDocument):
    """Container for non-conformant documents.

    This container allows us to interact with these documents without going
    through the more rigorous validation.
    """

    id: DID
    context: Annotated[Any, Field(alias="@context")] = None  # noqa: F722
    also_known_as: Any = None
    controller: Any = None
    verification_method: Any = None
    authentication: Any = None
    assertion_method: Any = None
    key_agreement: Any = None
    capability_invocation: Any = None
    capability_delegation: Any = None
    service: Any = None

    def _index_resources(self):
        """Index resources by ID.

        This is done in the most permissive way possible. ID collisions will
        result in overwritten data instead of raising an error.
        """

        def _indexer(item):
            if isinstance(item, list):
                # Recurse on lists
                for subitem in item:
                    _indexer(subitem)
                return

            if not isinstance(item, dict):
                # Only dictionaries are indexable
                return

            if "id" not in item:
                # Only dictionaries with IDs are indexed
                return

            # Attempt to account for relative IDs
            try:
                ref = DIDUrl(item["id"])
                if not ref.did:
                    key = ref.as_absolute(self.id)
                else:
                    key = ref
            except (InvalidDIDError, InvalidDIDUrlError):
                key = item["id"]

            self._index[key] = Resource(**item)

            # Recurse
            for value in item.values():
                _indexer(value)

        for _, value in self:
            _indexer(value)

    def dereference(self, reference: Union[str, DIDUrl]) -> Resource:
        """Dereference a DID URL to a document resource."""
        if isinstance(reference, str):
            reference = DIDUrl.parse(reference)
        if not reference.did:
            reference = reference.as_absolute(self.id)

        if reference not in self._index:
            raise IDNotFoundError("ID {} not found in document".format(reference))
        return self._index[reference]



"""DID Document classes."""

from .builder import (
    DIDDocumentBuilder,
    RelationshipBuilder,
    ServiceBuilder,
    VerificationMethodBuilder,
)
from .doc import (
    BasicDIDDocument,
    DIDDocument,
    DIDDocumentError,
    DIDDocumentRoot,
    IdentifiedResourceMismatch,
    IDNotFoundError,
)

__all__ = [
    "DIDDocumentError",
    "IdentifiedResourceMismatch",
    "IDNotFoundError",
    "DIDDocumentRoot",
    "BasicDIDDocument",
    "DIDDocument",
    "VerificationMethodBuilder",
    "RelationshipBuilder",
    "ServiceBuilder",
    "DIDDocumentBuilder",
]



"""DID Doc Service."""

from typing import Any, List, Mapping, Optional, Union

from pydantic import AnyUrl, ConfigDict, StrictStr
from typing_extensions import Literal

from .did import DID
from .did_url import DIDUrl
from .resource import Resource

EndpointStrings = Union[DID, DIDUrl, AnyUrl, StrictStr]


class Service(Resource):
    """Representation of DID Document Services."""

    id: DIDUrl
    type: Union[str, List[str]]
    service_endpoint: Union[
        EndpointStrings,
        List[Union[EndpointStrings, Mapping[str, Any]]],
        Mapping[str, Any],
    ]


class DIDCommV1Service(Service):
    """DID Communication Service."""

    model_config = ConfigDict(extra="forbid")

    type: Literal["IndyAgent", "did-communication", "DIDCommMessaging"] = (
        "did-communication"
    )
    service_endpoint: EndpointStrings
    recipient_keys: List[DIDUrl]
    routing_keys: List[DIDUrl] = []
    accept: Optional[List[str]] = None
    priority: int = 0


DIDCommService = DIDCommV1Service


class DIDCommV2ServiceEndpoint(Resource):
    """DID Communication Service Endpoint."""

    uri: EndpointStrings
    accept: Optional[List[str]] = None
    routing_keys: List[DIDUrl] = []


class DIDCommV2Service(Service):
    """DID Communication V2 Service."""

    model_config = ConfigDict(extra="forbid")

    type: Literal["DIDCommMessaging"] = "DIDCommMessaging"
    service_endpoint: Union[List[DIDCommV2ServiceEndpoint], DIDCommV2ServiceEndpoint]


class UnknownService(Service):
    """Unknown Service."""



"""Generic DID Document Classes.

If using Python 3.7+, these can be used to simplify subclassing a DID Document.
These classes are experimental and subject to change without notice. Use at
your own risk.
"""

import sys
from typing import List, Optional, TypeVar, Union

from pydantic import BaseModel

from ..did_url import DIDUrl
from ..service import Service
from ..verification_method import VerificationMethod
from .doc import BasicDIDDocument, DIDDocumentRoot

if sys.version_info >= (3, 7):  # pragma: no cover
    from typing import Generic

    VM = TypeVar("VM", bound=VerificationMethod)
    SV = TypeVar("SV", bound=Service)
    Methods = Optional[List[VM]]
    Relationships = Optional[List[Union[DIDUrl, VM]]]
    Services = Optional[List[SV]]

    class GenericDIDDocumentRoot(DIDDocumentRoot, BaseModel, Generic[VM, SV]):
        """DID Document Root with Generics."""

        verification_method: Methods[VM] = None
        authentication: Relationships[VM] = None
        assertion_method: Relationships[VM] = None
        key_agreement: Relationships[VM] = None
        capability_invocation: Relationships[VM] = None
        capability_delegation: Relationships[VM] = None
        service: Services[SV] = None

    class GenericBasicDIDDocument(BasicDIDDocument, BaseModel, Generic[VM, SV]):
        """BasicDIDDocument with Generics."""

        verification_method: Methods[VM] = None
        authentication: Relationships[VM] = None
        assertion_method: Relationships[VM] = None
        key_agreement: Relationships[VM] = None
        capability_invocation: Relationships[VM] = None
        capability_delegation: Relationships[VM] = None
        service: Services[SV] = None



"""Common components."""

import re

DID_REGEX = "did:([a-z0-9]+):((?:[a-zA-Z0-9._%-]*:)*[a-zA-Z0-9._%-]+)"
DID_PATTERN = re.compile(f"^{DID_REGEX}$")
DID_URL_DID_PART_PATTERN = re.compile(f"^({DID_REGEX})[?/#]")
DID_URL_RELATIVE_FRONT = re.compile("^[?/#].*")


class DIDError(Exception):
    """General did error."""



"""PyDID."""

import logging
from typing import Callable, List, Optional, Type

from .common import DIDError
from .did import DID, InvalidDIDError
from .did_url import DIDUrl, InvalidDIDUrlError
from .doc import corrections, generic
from .doc.builder import DIDDocumentBuilder
from .doc.doc import (
    BaseDIDDocument,
    BasicDIDDocument,
    DIDDocument,
    DIDDocumentError,
    NonconformantDocument,
)
from .resource import Resource
from .service import DIDCommService, Service
from .verification_method import (
    VerificationMaterial,
    VerificationMaterialUnknown,
    VerificationMethod,
)

__all__ = [
    "BasicDIDDocument",
    "DID",
    "DIDCommService",
    "DIDDocument",
    "DIDDocumentBuilder",
    "DIDDocumentError",
    "DIDError",
    "DIDUrl",
    "InvalidDIDError",
    "InvalidDIDUrlError",
    "Service",
    "VerificationMethod",
    "VerificationMaterial",
    "VerificationMaterialUnknown",
    "Resource",
    "generic",
    "corrections",
]


LOGGER = logging.getLogger(__name__)


def deserialize_document(
    value: dict,
    corrections: Optional[List[Callable]] = None,
    *,
    strict: bool = False,
    cls: Optional[Type[BaseDIDDocument]] = None,
) -> BaseDIDDocument:
    """Deserialize a document from a dictionary."""
    if corrections:
        for correction in corrections:
            value = correction(value)

    cls = cls or DIDDocument
    if strict:
        return cls.deserialize(value)
    try:
        return cls.deserialize(value)
    except ValueError as error:
        LOGGER.warning("Failed to deserialize document: %s", error)
        LOGGER.info("Parsing document as non-conformant doc")

    return NonconformantDocument.deserialize(value)



"""DID URL Object."""

from typing import TYPE_CHECKING, Any, Dict, Optional
from urllib.parse import parse_qsl, urlencode, urlparse

from pydantic import GetCoreSchemaHandler, GetJsonSchemaHandler
from pydantic.json_schema import JsonSchemaValue
from pydantic_core import CoreSchema, core_schema

from .common import DID_URL_DID_PART_PATTERN, DID_URL_RELATIVE_FRONT, DIDError

if TYPE_CHECKING:  # pragma: no cover
    from .did import DID


class InvalidDIDUrlError(DIDError, ValueError):
    """Invalid DID."""


class DIDUrl(str):
    """DID URL."""

    def __init__(self, url: str):
        """Parse DID URL from string."""
        super().__init__()
        matches = DID_URL_DID_PART_PATTERN.match(url)

        if matches:
            self.did = matches.group(1)
            _, url_component = url.split(self.did)
        else:
            relative_matches = DID_URL_RELATIVE_FRONT.match(url)
            if not relative_matches:
                raise InvalidDIDUrlError(
                    "{} is not a valid absolute or relative DID URL".format(url)
                )
            self.did = None
            url_component = url

        parts = urlparse(url_component)
        self.path = parts.path or None
        self.query = dict(parse_qsl(parts.query)) if parts.query else None
        self.fragment = parts.fragment or None

    @classmethod
    def __get_pydantic_core_schema__(
        cls, source_type: Any, handler: GetCoreSchemaHandler
    ) -> CoreSchema:
        """Get core schema."""
        return core_schema.no_info_after_validator_function(cls, handler(str))

    @classmethod
    def __get_pydantic_json_schema__(
        cls, core_schema: CoreSchema, handler: GetJsonSchemaHandler
    ) -> JsonSchemaValue:
        """Update schema fields."""
        json_schema = handler(core_schema)
        json_schema["examples"] = ["did:example:123/some/path?query=test#fragment"]
        return json_schema

    @classmethod
    def parse(cls, url: str):
        """Parse DID URL from string."""
        return cls(url)

    @classmethod
    def unparse(
        cls,
        did: str,
        path: Optional[str] = None,
        query: Optional[Dict[str, str]] = None,
        fragment: Optional[str] = None,
    ) -> "DIDUrl":
        """Form a DID URL from parts and return the string representation."""
        value = did
        if path and not path.startswith("/"):
            path = "/" + path

        if path:
            value += path

        if query:
            value += "?" + urlencode(query)

        fragment = str(fragment) if fragment else None
        if fragment:
            value += "#" + fragment

        return cls(value)

    def as_absolute(self, did: "DID"):
        """Make a relative DIDUrl absolute."""
        return self.unparse(did, self.path, self.query, self.fragment)

    @classmethod
    def is_valid(cls, url: str):
        """Return whether the given string is a valid DID URL."""
        try:
            cls.validate(url)
        except ValueError:
            return False
        else:
            return True

    @classmethod
    def validate(cls, url: str):
        """Validate the given url as a DID URL."""
        return cls.parse(url)
